/*
 * Copyright (C) 2017 Open Source Robotics Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
*/

//                ****** Do not modify this file. ******
// This file is automatically generated by CMake. Changes should instead be
// made to cmake/gz_auto_headers.hh.in in gz-cmake

#include <gz/utils/config.hh>
#include <gz/utils/cli/App.hpp>
#include <gz/utils/cli/CLI.hpp>
#include <gz/utils/cli/Config.hpp>
#include <gz/utils/cli/ConfigFwd.hpp>
#include <gz/utils/cli/Error.hpp>
#include <gz/utils/cli/Formatter.hpp>
#include <gz/utils/cli/FormatterFwd.hpp>
#include <gz/utils/cli/Macros.hpp>
#include <gz/utils/cli/Option.hpp>
#include <gz/utils/cli/Split.hpp>
#include <gz/utils/cli/StringTools.hpp>
#include <gz/utils/cli/Timer.hpp>
#include <gz/utils/cli/TypeTools.hpp>
#include <gz/utils/cli/Validators.hpp>
#include <gz/utils/cli/Version.hpp>

