/*
 * Copyright (C) 2022 Open Source Robotics Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

/* Config.hh. Generated by CMake for gz-utils. */

#ifndef GZ_UTILS_CONFIG_HH_
#define GZ_UTILS_CONFIG_HH_

/* Version number */
#define GZ_UTILS_MAJOR_VERSION 2
#define GZ_UTILS_MINOR_VERSION 2
#define GZ_UTILS_PATCH_VERSION 1

#define GZ_UTILS_VERSION "2.2"
#define GZ_UTILS_VERSION_FULL "2.2.1"

#define GZ_UTILS_VERSION_NAMESPACE v2

#define GZ_UTILS_VERSION_HEADER "Gazebo Utils, version 2.2.1\nCopyright (C) 2020 Open Source Robotics Foundation.\nReleased under the Apache 2.0 License.\n\n"

#endif
